'''
Generate a YAML file with the pixel layout and chip connections for the
28-chip sensor plane, with no focusing grids or extended pads, for 8
chips in one column.

'''
import yaml
import patterngenerator as pg

pixels = []
subgrid_width = 12
pixels.extend(pg.pixels_triangle_grid(subgrid_width, 4, 1, 76, 52, 0))
pixels.extend(pg.pixels_triangle_grid(subgrid_width, 6, 1, 64, 64,
    len(pixels)))
pixels.extend(pg.pixels_triangle_grid(subgrid_width, 8, 2, 52, 76,
    len(pixels)))
pixels.extend(pg.pixels_plain_grid(3, 8, 2, 53.5, 101.5, len(pixels)))
pixels.extend(pg.pixels_plain_grid(3, 6, 1, 65.5, 125.5, len(pixels)))
pixels.extend(pg.pixels_plain_grid(3, 4, 1, 77.5, 137.5, len(pixels)))

pixelids = {
        # Bool value is argument to right_side_up
        # ranges are pixel id ranges for 0-15, 16-31
        48: (False, 'triangle', range(1*16, 2*16), range(0*16, 1*16)),
        51: (False, 'triangle', range(6*16, 7*16), range(5*16, 6*16)),
        53: (False, 'triangle', range(13*16, 14*16), range(12*16, 13*16)),
        57: (False, 'triangle', range(21*16, 22*16), range(20*16, 21*16)),
        54: (False, 'plain', range(29*16, 30*16), range(28*16, 29*16)),
        58: (False, 'plain', range(37*16, 38*16), range(36*16, 37*16)),
        60: (False, 'plain', range(44*16, 45*16), range(43*16, 44*16)),
        63: (False, 'plain', range(49*16, 50*16), range(48*16, 49*16)),

}

chips = []
for chipid, (right_side_up, shape, first_ids, second_ids) in pixelids.items():
    if shape == 'plain':
        assignment = pg.grid_4x4_assignments
    else:
        assignment = pg.triangle_assignments
    first_channels = pg.assign_pixels(first_ids, assignment, right_side_up)
    second_channels = pg.assign_pixels(second_ids, assignment, right_side_up)
    chips.append([chipid, first_channels + second_channels])


with open('sensor_plane_28_8chip.yaml', 'w') as f:
    yaml.dump({'pixels': pixels, 'chips': chips}, f)



'''
Geometry of the LArPix pixel plane.

'''

class PixelPlane(object):
    '''
    The pixel plane for LArPix including pixel pads and LArPix chips.

    '''
    def __init__(self):
        self.pixels = {}
        self.chips = {}
        self.unconnected_pixel = Pixel()
        self.unconnected_pixel.channel_connection = []

    @classmethod
    def fromDict(cls, d):
        '''
        Create a new pixel plane using the data in the dict.

        Dict format:

        >>> {'pixels': [(pixelid, x, y, [(pad_vertex_1_x, y), ...],
        ...     [(focus_vertex_1_x, y), ...]), ...],  # list of pixels
        ...  'chips': [(chipid, [ch1_pixelid, ...,]), ...]  # list of chips
        ... }

        If the sensor pad is just a via, pass an empty list for the pad
        vertices. If there is no focusing grid, pass an empty list for
        the focus vertices.

        '''
        result = cls()
        for i, (pixelid, x, y, pad_outline, focus_outline) in enumerate(d['pixels']):
            pixel = Pixel()
            pixel.pixelid = pixelid
            pixel.x = x
            pixel.y = y
            pixel.pad_outline = pad_outline
            pixel.focus_outline = focus_outline
            result.pixels[pixelid] = pixel
        for i, (chipid, channel_connections) in enumerate(d['chips']):
            chip = GeomChip()
            chip.chipid = chipid
            for channel, pixelid in enumerate(channel_connections):
                if pixelid == None:
                    chip.channel_connections.append(result.unconnected_pixel)
                    result.unconnected_pixel.channel_connection.append((chip,
                        channel))
                else:
                    chip.channel_connections.append(result.pixels[pixelid])
                    result.pixels[pixelid].channel_connection = (chip, channel)
            result.chips[chipid] = chip
        return result

    def channels_where(self, condition):
        '''
        Return a list of (chip, channel) for the pixels that satisfy the
        given condition.

        ``condition`` should be a function of one argument, a ``Pixel``
        object.

        >>> pixelplane.channels_where(lambda pixel: pixel.x > 30 and not pixel.focus_outline)

        '''
        good_pixels = filter(condition, self.pixels.values())
        return [pixel.channel_connection for pixel in good_pixels]


class GeomChip(object):
    '''
    A LArPix chip to associate with geometric features.

    Note: Channels which are not connected to any pixel have, by
    convention, are assigned to a pixel index of None.

    '''
    def __init__(self):
        self.channel_connections = []
        self.chipid = None

class Pixel(object):
    '''
    A pixel pad + focusing region on the LArPix pixel plane.

    '''
    def __init__(self):
        self.pixelid = None
        self.x = 0
        self.y = 0
        self.pad_outline = []
        self.focus_outline = []
        self.channel_connection = None



'''
Generate pixel plane patterns.

'''

from itertools import product
import numpy as np

def pixels_plain_grid(pixel_pitch, nblocksx, nblocksy, startx, starty, start_index):
    '''
    A plain grid of no-pad no-focus pixels, numbered in batches of 4x4.

    '''
    pixels = []
    pixels_per_grid = 16
    repetition_period = 4*pixel_pitch
    x, y, = np.meshgrid(pixel_pitch*np.arange(4),
            pixel_pitch*np.arange(4))
    subgrid = np.array(list(zip(x.reshape(-1), y.reshape(-1))))
    # This line forms the list of blocks going along rows. Without this
    # nonsense (just using product(range, range) goes down columns).
    blocklist = np.array(list(product(range(nblocksy), range(nblocksx))))[:, ::-1]
    for block_index, (xblock, yblock) in enumerate(blocklist):
        pixelids = range(block_index*pixels_per_grid + start_index,
                (block_index+1)*pixels_per_grid + start_index)
        offset = np.array([xblock*repetition_period + startx,
            yblock*repetition_period + starty])
        pixel_locations = subgrid + offset
        for pixelid, (x, y) in zip(pixelids, pixel_locations):
            pixels.append([pixelid, float(x), float(y), [], []])
    return pixels

def pixels_triangle_grid(repetition_period, nblocksx, nblocksy, startx,
        starty, start_index):
    '''
    A grid of triangle pixels specific to the LArPix sensor board.

    '''
    pixels = []
    unit = repetition_period / 8.0
    subgrid = np.array(  # Laid out here in the orientation on the board
              [[2*unit, unit],                [6*unit, unit],
        [unit, 2*unit], [3*unit, 2*unit], [5*unit, 2*unit], [7*unit, 2*unit],
               [2*unit, 10/3.*unit],          [6*unit, 10/3.*unit],
               [2*unit, 14/3.*unit],          [6*unit, 14/3.*unit],
        [unit, 6*unit], [3*unit, 6*unit], [5*unit, 6*unit], [7*unit, 6*unit],
               [2*unit, 7*unit],              [6*unit, 7*unit]])
    pixels_per_grid = len(subgrid)
    # This line forms the list of blocks going along rows. Without this
    # nonsense (just using product(range, range) goes down columns).
    blocklist = np.array(list(product(range(nblocksy), range(nblocksx))))[:, ::-1]
    for block_index, (xblock, yblock) in enumerate(blocklist):
        pixelids = range(block_index*pixels_per_grid + start_index,
                (block_index+1)*pixels_per_grid + start_index)
        offset = np.array([xblock*repetition_period + startx,
            yblock*repetition_period + starty])
        pixel_locations = subgrid + offset
        for pixelid, (x, y) in zip(pixelids, pixel_locations):
            pixels.append([pixelid, float(x), float(y), [], []])
    return pixels

grid_4x4_assignments = [14, 13, 15, 12, 10, 9, 11, 8, 7, 4, 6, 5, 3, 0, 2, 1]
'''
Index in flattened ``pixelids`` list (from the chip side)::

0  1  2  3
4  5  6  7
8  9  10 11
12 13 14 15

Channel assignments (right side up) (from the chip side)::

13 15 14 12     29 31 30 28
9  11 10 8  or  25 27 26 24
7  5  4  6      23 21 20 22
3  1  0  2      19 17 16 18

'''

triangle_assignments = [15, 14, 12, 11, 13, 10, 9, 8, 7, 6, 5, 2, 4, 3, 1, 0]
'''
Index in flattened ``pixelids`` list (from the chip side)::

   0           1
2     3     4     5
   6           7
   8           9
10    11    12    13
   14          15

Channel assignments (right-side up) (from the chip side)::

   15          14              31          30
11    13    12    10        27    29    28    26
   9           8       or      25          24
   7           6               23          22
5     3     2     4         21    19    18    20
   1           0               17          16

'''

def assign_pixels(pixelids, assignments, right_side_up=True):
    '''
    Return a list assigning the given pixelids to a chip's channels
    using the provided assignment array.

    The right-side up configuration (``right_side_up=True``) of the chip
    is (viewed from above), channels 0-15 on the left and channels 16-31
    on the right. The upside-down configuration is reversed: channels
    16-31 on the left, and channels 0-15 on the right.

    Since the convention is that channel id is given by position in the
    list, this function always returns assignments to channels 0-15.
    If the given grid should be assigned to channels 16-31, just
    concatenate it with the assignments for 0-15.

    ``assignments[i]`` gives the index in the final list of
    ``pixelids[i]``.

    '''
    if not right_side_up:
        assignments = assignments[::-1]
    channel_connections = []
    for channel_id, pixelid in enumerate(pixelids):
        channel_connections.append(pixelids[assignments[channel_id]])
    return channel_connections



'''
Generate a YAML file with the pixel layout and chip connections for the
28-chip sensor plane, with no focusing grids or extended pads, for 8
chips in one column.

'''
import yaml
import patterngenerator as pg

pixels = []
subgrid_width = 12
pixels.extend(pg.pixels_triangle_grid(subgrid_width, 4, 1, 76, 52, 0))
pixels.extend(pg.pixels_triangle_grid(subgrid_width, 6, 1, 64, 64,
    len(pixels)))
pixels.extend(pg.pixels_triangle_grid(subgrid_width, 8, 2, 52, 76,
    len(pixels)))
pixels.extend(pg.pixels_plain_grid(3, 8, 2, 53.5, 101.5, len(pixels)))
pixels.extend(pg.pixels_plain_grid(3, 6, 1, 65.5, 125.5, len(pixels)))
pixels.extend(pg.pixels_plain_grid(3, 4, 1, 77.5, 137.5, len(pixels)))

pixelids = {
        # Bool value is argument to right_side_up
        # ranges are pixel id ranges for 0-15, 16-31
        3: (True, 'triangle', [None]*16, range(4*16, 5*16)),
        5: (True, 'triangle', range(10*16, 11*16), range(11*16, 12*16)),
        6: (True, 'triangle', range(18*16, 19*16), range(19*16, 20*16)),
        9: (True, 'plain', range(26*16, 27*16), range(27*16, 28*16)),
        10: (True, 'plain', range(34*16, 35*16), range(35*16, 36*16)),
        12: (True, 'plain', [None]*16, range(42*16, 43*16)),
        48: (False, 'triangle', range(1*16, 2*16), range(0*16, 1*16)),
        51: (False, 'triangle', range(6*16, 7*16), range(5*16, 6*16)),
        53: (False, 'triangle', range(13*16, 14*16), range(12*16, 13*16)),
        57: (False, 'triangle', range(21*16, 22*16), range(20*16, 21*16)),
        54: (False, 'plain', range(29*16, 30*16), range(28*16, 29*16)),
        58: (False, 'plain', range(37*16, 38*16), range(36*16, 37*16)),
        60: (False, 'plain', range(44*16, 45*16), range(43*16, 44*16)),
        63: (False, 'plain', range(49*16, 50*16), range(48*16, 49*16)),
        80: (True, 'triangle', range(2*16, 3*16), range(3*16, 4*16)),
        83: (True, 'triangle', range(7*16, 8*16), range(8*16, 9*16)),
        85: (True, 'triangle', range(14*16, 15*16), range(15*16, 16*16)),
        86: (True, 'triangle', range(22*16, 23*16), range(23*16, 24*16)),
        89: (True, 'plain', range(30*16, 31*16), range(31*16, 32*16)),
        90: (True, 'plain', range(38*16, 39*16), range(39*16, 40*16)),
        92: (True, 'plain', range(45*16, 46*16), range(46*16, 47*16)),
        95: (True, 'plain', range(50*16, 51*16), range(51*16, 52*16)),
        108: (False, 'triangle', [None]*16, range(9*16, 10*16)),
        106: (False, 'triangle', range(17*16, 18*16), range(16*16, 17*16)),
        105: (False, 'triangle', range(25*16, 26*16), range(24*16, 25*16)),
        102: (False, 'plain', range(33*16, 34*16), range(32*16, 33*16)),
        101: (False, 'plain', range(41*16, 42*16), range(40*16, 41*16)),
        99: (False, 'plain', [None]*16, range(47*16, 48*16)),


}

chips = []
for chipid, (right_side_up, shape, first_ids, second_ids) in pixelids.items():
    if shape == 'plain':
        assignment = pg.grid_4x4_assignments
    else:
        assignment = pg.triangle_assignments
    first_channels = pg.assign_pixels(first_ids, assignment, right_side_up)
    second_channels = pg.assign_pixels(second_ids, assignment, right_side_up)
    chips.append([chipid, first_channels + second_channels])


with open('sensor_plane_28_full.yaml', 'w') as f:
    yaml.dump({'pixels': pixels, 'chips': chips}, f)



import yaml
import os

def load(filename):
    '''
    Load the specified layout file.

    The path is first searched relative to the "current directory".
    If no match is found, the path is searched relative to the "config"
    package directory (aka ``__file__`` directly in code).

    '''
    if os.path.isfile(filename):
        with open(filename, 'r') as f:
            return yaml.load(f)
    elif os.path.isfile(os.path.join(os.path.dirname(__file__), filename)):
        with open(os.path.join(os.path.dirname(__file__), filename), 'r') as f:
            return yaml.load(f)
    else:
        raise IOError('File not found: %s' % filename)



'''
Generate a PDF of a pixel plane YAML file.

'''
from reportlab.pdfgen import canvas
from reportlab.lib.pagesizes import inch
import yaml
from larpixgeometry.pixelplane import PixelPlane
import  numpy as np

with open('sensor_plane_28_full.yaml', 'r') as f:
    pixelplane = PixelPlane.fromDict(yaml.load(f))

colors = np.array([[228, 26, 28], [55, 126, 184], [77, 175, 74], [152,
    78, 163], [255, 127, 0]])/256.0
colors = np.tile(colors, (10,1))



c = canvas.Canvas('sensor_28chip.pdf', pagesize=(12*inch, 12*inch))
c.setFont('Helvetica', 20)
c.drawString(4*inch, 11*inch, 'Sensor board with 28 chips')
c.drawString(4*inch, 10.6*inch, '(view from chip side)')
c.setFont('Courier', 7)
colorkey = []
for pixel in pixelplane.pixels.values():
    c.circle(pixel.x*7-300, 15*inch-pixel.y*7, 0.4)
    c.drawCentredString(pixel.x*7-300, 15*inch-pixel.y*7, str(pixel.pixelid))
for chip, color in zip(pixelplane.chips.values(), colors):
    c.setFont('Courier-Bold', 11)
    c.setFillColorRGB(*color, alpha=1)
    colorkey.append(['Chip %d' % chip.chipid, color])
    x_sum = 0
    y_sum = 0
    count = 0
    for channel, pixel in enumerate(chip.channel_connections):
        if pixel != pixelplane.unconnected_pixel:
            c.drawCentredString(pixel.x*7-300, 15*inch-pixel.y*7-7,
                    str(channel))
            x_sum += pixel.x*7-300
            y_sum += 15*inch-pixel.y*7 - 5
            count += 1
    x_avg = x_sum/float(count)
    y_avg = y_sum/float(count)
    font = ('Helvetica', 56)
    c.setFont(*font)
    c.setFillColorRGB(*color, alpha=0.45)
    width = c.stringWidth(str(chip.chipid), *font)
    c.drawCentredString(x_avg, y_avg-16, str(chip.chipid))
# for i, (string, color) in enumerate(colorkey):
    # c.setFillColorRGB(*color)
    # c.drawString(80, 11*inch-25*i, string)
c.showPage()
c.save()



'''
The setup.py file for larpix-geometry.

'''

from setuptools import setup, find_packages

setup(
        name='larpix-geometry',
        version='0.3.0',
        description='LArPix sensor plane geometry',
        url='https://github.com/samkohn/larpix-geometry',
        author='Sam Kohn',
        author_email='skohn@lbl.gov',
        keywords='dune physics',
        packages=find_packages(),
        install_requires=['pyyaml'],
        package_data={
            'larpixgeometry.layouts':['*.yaml']
        },
)
